/**************************************************************************/
/**
**  @file
**  @brief prototype of meander_
*/
/**************************************************************************/

#if !defined(MEANDER_H_INCLUDED)
#define MEANDER_H_INCLUDED

void meander_( const int *, const int *, const double *, const double *,
	       const double *, const double *, 
	       const double *, const double *, const double *,
	       const double *, const double *, 
	       const double *, const double *, double *, double *, double *,
	       double *, double * ); 

#endif
