/*************************************************************************/
/**
**  @file tMatrix.cpp
**  @brief Definition of class tMatrix.
**
**  Class tMatrix implements matrices (2D arrays) as 1D arrays of tArray
**  objects. The () operator is overloaded so that you can access array
**  elements in the "old fashioned" way, like "mymatrix( 42, 24 );".
**  One of the constructors takes two integer arguments representing the
**  size of the matrix.
**
**  $Id: tMatrix.cpp,v 1.6 2004-02-27 18:43:37 childcvs Exp $
*/
/*************************************************************************/

#include "tMatrix.h"

